##################################################################
# NYISO DCR Reset 2020: Net EAS Revenues Model
# Helper Functions
#_________________________________________________________________

# Script purpose 
#   This script contains a series of functions used in various scripts throughout the model for importing and exporting data and for handling daylight-savings time adjustments.

#### xlsx_import #####

xlsx_import <- function(source, sht) {
  import_file <- read_xlsx(source
                           , sheet = sht
                           , trim_ws = TRUE
                           , .name_repair = "universal")
  if (dim(import_file[names(import_file) == "Date"])[2] == 1) {
    import_file$Date <- as_date(import_file$Date)
    return(import_file)
  }
  else {
    return(import_file)
  }
}

#### unit_import ####

unit_import <- function(case_num, all_cases) {
  case_to_setup <- all_cases[all_cases$Case_Number == case_num,]
  last_unit <- as.name(paste0("Unit_", case_to_setup$Number_of_Units))
  unit_list <- as.character(case_to_setup[1, 5:eval(5+case_to_setup$Number_of_Units-1)])
  toggles <- select(case_to_setup, Unit_Type:Case_ID, SCR_Toggle:last_col())
  stop_flag <- 0
  if("j15" %in% unit_list & toggles$SCR_Toggle == "Yes"){
    warning <- c("Be aware that the model does not contain parameters for with-SCR operation for the H-class 15ppm unit. The model will run but this unit will not be evaluated.")
    unit_list <- unit_list[which(unit_list != "j15")]
  }
  if("ccj" %in% unit_list & toggles$Unit_Type == "CT"){
    warning <- c("In at least one of your cases, you have selected your unit type as CT but your unit as a combined-cycle unit (ccj). The model will not run.")
    stop_flag <- 1
  }
  if((TRUE %in% unique(c("j25", "j15", "f", "aero") %in% unit_list)) & toggles$Unit_Type == "CC"){
    warning <- c("In at least one of your cases, you have selected your unit type as CC but at least one of your units is not a combined-cycle unit. The model will not run.")
    stop_flag <- 1
  }
  if("ccj" %in% unit_list & toggles$SCR_Toggle == "No"){
    warning <- c("In at least one of your cases, you have selected no-SCR operation for the H-class combined cycle unit. The model does not support this and will not run.")
    stop_flag <- 1
  }
  if(stop_flag == 1){
    stop(warning)
  }
  case_list <- list(unit_list, toggles)
  names(case_list) <- c("Unit_List", "Toggles")
  return(case_list)
}

#### dls_adjust ####

dls_adjust <- function(price_set) {
  dup_vector <- duplicated(select(price_set, DateTime, Zone))
  price_set <- data.frame(price_set, dup_vector)
  price_set$DateTime[price_set$dup_vector == TRUE] <- price_set$DateTime[price_set$dup_vector == TRUE] + seconds(1)
  dup_check <- duplicated(select(price_set, DateTime, Zone))
  ifelse(sum(dup_check) > 0, stop("Still Duplicates Remaining"), return(select(price_set, -dup_vector)))
}

#### exp_datasets and exp_csv ####

exp_datasets <- function(int_output_toggle, format, datasets, exp_name, case_vec) {
  if(int_output_toggle == "No"){
    print("Execution instructions specify no output of intermediate datasets")
  }else{
    if (format == ".rds") {
      exp_pathname <- paste0(intermediate_datasets, exp_name)
      if(.Platform$OS.type == "windows" & str_length(exp_pathname) > 251) {
        warning("Truncating intermediate dataset name to prevent filepath from exceeding maximum Windows limit.")
        exp_pathname <- str_sub(exp_pathname, end = 251)
      }
      write_rds(datasets, path = paste0(exp_pathname, ".rds"), compress = c("gz"))  
    }else if(format == ".csv"){
      lapply(case_vec, exp_csv, datasets = datasets, exp_name = exp_name)
    }else if(format == "Both (.rds, .csv)"){
      exp_pathname <- paste0(intermediate_datasets, exp_name)
      if(.Platform$OS.type == "windows" & str_length(exp_pathname) > 251) {
        warning("Truncating intermediate dataset name to prevent filepath from exceeding maximum Windows limit.")
        exp_pathname <- str_sub(exp_pathname, end = 251)
      }
      write_rds(datasets, path = paste0(exp_pathname, ".rds"), compress = c("gz"))  
      lapply(case_vec, exp_csv, datasets = datasets, exp_name = exp_name)
    }else{
      stop("Format specified incorrectly. Must be either .rds or .csv if Generate_Intermediate_Output is set to Yes.")
    }
    return(paste("See Intermediate Datasets folder for output:", intermediate_datasets))
  }
}

exp_csv <- function(case_to_exp, datasets, exp_name, zone_info) {
  case <- paste0("Case_", case_to_exp)
  dfs_to_exp <- datasets[[case]][[2]]
  toggles <- datasets[[case]][[1]][[2]]
  exp_pathname <- paste0(intermediate_datasets, paste(exp_name, 
                                                      toggles$Case_ID))
  if(.Platform$OS.type == "windows" & str_length(exp_pathname) > 245) {
    warning("Truncating intermediate dataset name to prevent filepath from exceeding maximum Windows limit.")
    exp_pathname <- str_sub(exp_pathname, end = 245)
  }
  if(is.data.frame(dfs_to_exp)){
    invisible(write_csv(dfs_to_exp, path = paste0(exp_pathname, ".csv")))  
  }else{
    dataset_list <- names(dfs_to_exp)
    invisible(lapply(dataset_list, function(x) write_csv(dfs_to_exp[[x]], path = paste0(exp_pathname, " ", tolower(x), ".csv"))))
  }
}
